@@ -10,14 +10,18 @@
 """
 
 import codecs
-import sha
+try:
+    from hashlib import sha1
+except ImportError:
+    from sha import new as sha1
 import time
 from os import path
 from threading import Lock
 from jinja.parser import Parser
 from jinja.translators.python import PythonTranslator, Template
-from jinja.exceptions import TemplateNotFound, TemplateSyntaxError
-from jinja.utils import CacheDict, raise_syntax_error
+from jinja.exceptions import TemplateNotFound, TemplateSyntaxError, \
+     TemplateIncludeError
+from jinja.utils import CacheDict
 
 
 #: when updating this, update the listing in the jinja package too
@@ -38,9 +42,8 @@ def get_cachename(cachepath, name, salt=None):
     Return the filename for a cached file.
     """
     return path.join(cachepath, 'jinja_%s.cache' %
-                     sha.new('jinja(%s|%s)tmpl' %
-                             (name, salt or '')).hexdigest())
-
+                     sha1('jinja(%s|%s)tmpl' %
+                          (name, salt or '')).hexdigest())
 
 
 def _loader_missing(*args, **kwargs):
@@ -48,7 +51,6 @@ def _loader_missing(*args, **kwargs):
     raise RuntimeError('no loader defined')
 
 
-
 class LoaderWrapper(object):
     """
     Wraps a loader so that it's bound to an environment.
@@ -96,8 +98,16 @@ class LoaderWrapper(object):
             if not self.environment.friendly_traceback:
                 raise
             __traceback_hide__ = True
+            from jinja.debugger import raise_syntax_error
             raise_syntax_error(e, self.environment)
 
+    def get_controlled_loader(self):
+        """
+        Return a loader that runs in a controlled environment.  (Keeps
+        track of templates that it loads and is not thread safe).
+        """
+        return ControlledLoader(self.environment, self.loader)
+
     def _loader_missing(self, *args, **kwargs):
         """Helper method that overrides all other methods if no
         loader is defined."""
@@ -107,6 +117,46 @@ class LoaderWrapper(object):
         return self.available
 
 
+class ControlledLoader(LoaderWrapper):
+    """
+    Used for template extending and including.
+    """
+
+    def __init__(self, environment, loader):
+        LoaderWrapper.__init__(self, environment, loader)
+        self._stack = []
+
+    def get_controlled_loader(self):
+        raise TypeError('Cannot get new controlled loader from an already '
+                        'controlled loader.')
+
+    def mark_as_processed(self):
+        """Mark the last parsed/sourced/included template as processed."""
+        if not self._stack:
+            raise RuntimeError('No template for marking found')
+        self._stack.pop()
+
+    def _controlled(method):
+        def new_method(self, name, *args, **kw):
+            if name in self._stack:
+                raise TemplateIncludeError('Circular imports/extends '
+                                           'detected.  %r appeared twice.' %
+                                           name)
+            self._stack.append(name)
+            return method(self, name, *args, **kw)
+        try:
+            new_method.__name__ = method.__name__
+            new_method.__doc__ = method.__doc__
+        except AttributeError:
+            pass
+        return new_method
+
+    get_source = _controlled(LoaderWrapper.get_source)
+    parse = _controlled(LoaderWrapper.parse)
+    load = _controlled(LoaderWrapper.load)
+    del _controlled
+
+
 class BaseLoader(object):
     """
     Use this class to implement loaders.
@@ -264,22 +314,34 @@ class MemcachedLoaderMixin(object):
     """
     Uses a memcached server to cache the templates.
 
-    Requires the memcache library from tummy__.
+    Requires the memcache library from `tummy`_ or the cmemcache library
+    from `Gijsbert de Haan`_.
 
-    __ http://www.tummy.com/Community/software/python-memcached/
+    With Jinja 1.2 onwards you can also provide a `client` keyword argument
+    that takes an already instanciated memcache client or memcache client
+    like object.
+
+    .. _tummy: http://www.tummy.com/Community/software/python-memcached/
+    .. _Gisjsbert de Haan: http://gijsbert.org/cmemcache/
     """
 
     def __init__(self, use_memcache, memcache_time=60 * 60 * 24 * 7,
-                 memcache_host=None, item_prefix='template/'):
-        try:
-            from memcache import Client
-        except ImportError:
-            raise RuntimeError('the %r loader requires an installed '
-                               'memcache module' % self.__class__.__name__)
+                 memcache_host=None, item_prefix='template/', client=None):
         if memcache_host is None:
             memcache_host = ['127.0.0.1:11211']
         if use_memcache:
-            self.__memcache = Client(list(memcache_host))
+            if client is None:
+                try:
+                    try:
+                        from cmemcache import Client
+                    except ImportError:
+                        from memcache import Client
+                except ImportError:
+                    raise RuntimeError('the %r loader requires an installed '
+                                       'memcache module' %
+                                       self.__class__.__name__)
+                client = Client(list(memcache_host))
+            self.__memcache = client
             self.__memcache_time = memcache_time
         else:
             self.__memcache = None
@@ -472,24 +534,54 @@ class BasePackageLoader(BaseLoader):
 
     It uses the `pkg_resources` libraries distributed with setuptools for
     retrieving the data from the packages. This works for eggs too so you
-    don't have to mark your egg as non zip safe.
+    don't have to mark your egg as non zip safe. If pkg_resources is not
+    available it just falls back to path joining relative to the package.
     """
 
-    def __init__(self, package_name, package_path):
+    def __init__(self, package_name, package_path, force_native=False):
         try:
             import pkg_resources
         except ImportError:
             raise RuntimeError('setuptools not installed')
         self.package_name = package_name
         self.package_path = package_path
+        self.force_native = force_native
+
+    def _get_load_func(self):
+        if hasattr(self, '_load_func'):
+            return self._load_func
+        try:
+            from pkg_resources import resource_exists, resource_string
+            if self.force_native:
+                raise ImportError()
+        except ImportError:
+            basepath = path.dirname(__import__(self.package_name, None, None,
+                                    ['__file__']).__file__)
+            def load_func(name):
+                filename = path.join(basepath, *(
+                    self.package_path.split('/') +
+                    [p for p in name.split('/') if p != '..'])
+                )
+                if path.exists(filename):
+                    f = file(filename)
+                    try:
+                        return f.read()
+                    finally:
+                        f.close()
+        else:
+            def load_func(name):
+                path = '/'.join([self.package_path] + [p for p in name.split('/')
+                         if p != '..'])
+                if resource_exists(self.package_name, path):
+                    return resource_string(self.package_name, path)
+        self._load_func = load_func
+        return load_func
 
     def get_source(self, environment, name, parent):
-        from pkg_resources import resource_exists, resource_string
-        path = '/'.join([self.package_path] + [p for p in name.split('/')
-                         if p != '..'])
-        if not resource_exists(self.package_name, path):
+        load_func = self._get_load_func()
+        contents = load_func(name)
+        if contents is None:
             raise TemplateNotFound(name)
-        contents = resource_string(self.package_name, path)
         return contents.decode(environment.template_charset)
 
 
