@@ -15,9 +15,9 @@ import re
 import sys
 import string
 from types import MethodType, FunctionType
-from compiler.ast import CallFunc, Name, Const
-from jinja.nodes import Trans
+from jinja import nodes
 from jinja.exceptions import SecurityException, TemplateNotFound
+from jinja.datastructure import TemplateData
 
 # the python2.4 version of deque is missing the remove method
 # because a for loop with a lookup for the missing value written
@@ -61,13 +61,43 @@ except NameError:
 try:
     sorted = sorted
 except NameError:
-    def sorted(seq, reverse=False):
+    _cmp = cmp
+    def sorted(seq, cmp=None, key=None, reverse=False):
         rv = list(seq)
-        rv.sort()
+        if key is not None:
+            cmp = lambda a, b: _cmp(key(a), key(b))
+        rv.sort(cmp)
         if reverse:
             rv.reverse()
         return rv
 
+# group by support
+try:
+    from itertools import groupby
+except ImportError:
+    class groupby(object):
+
+        def __init__(self, iterable, key=lambda x: x):
+            self.keyfunc = key
+            self.it = iter(iterable)
+            self.tgtkey = self.currkey = self.currvalue = xrange(0)
+
+        def __iter__(self):
+            return self
+
+        def next(self):
+            while self.currkey == self.tgtkey:
+                self.currvalue = self.it.next()
+                self.currkey = self.keyfunc(self.currvalue)
+            self.tgtkey = self.currkey
+            return (self.currkey, self._grouper(self.tgtkey))
+
+        def _grouper(self, tgtkey):
+            while self.currkey == tgtkey:
+                yield self.currvalue
+                self.currvalue = self.it.next()
+                self.currkey = self.keyfunc(self.currvalue)
+
 #: function types
 callable_types = (FunctionType, MethodType)
 
@@ -167,30 +197,41 @@ def get_attribute(obj, name):
         raise AttributeError(name)
     if name[:2] == name[-2:] == '__':
         raise SecurityException('not allowed to access internal attributes')
-    if obj.__class__ in callable_types and name.startswith('func_') or \
-       name.startswith('im_'):
+    if getattr(obj, '__class__', None) in callable_types and \
+       name.startswith('func_') or name.startswith('im_'):
         raise SecurityException('not allowed to access function attributes')
     r = _getattr(obj, 'jinja_allowed_attributes', None)
     if r is not None and name not in r:
         raise SecurityException('disallowed attribute accessed')
-    return _getattr(obj, name)
+
+    # attribute lookups convert unicode strings to ascii bytestrings.
+    # this process could raise an UnicodeEncodeError.
+    try:
+        return _getattr(obj, name)
+    except UnicodeError:
+        raise AttributeError(name)
 
 
 def safe_range(start, stop=None, step=None):
     """
     "Safe" form of range that does not generate too large lists.
     """
-    # this also works with None since None is always smaller than
-    # any other value.
-    if start > MAX_RANGE:
-        start = MAX_RANGE
-    if stop > MAX_RANGE:
-        stop = MAX_RANGE
     if step is None:
         step = 1
     if stop is None:
-        return range(0, start, step)
-    return range(start, stop, step)
+        r = xrange(0, start, step)
+    else:
+        r = xrange(start, stop, step)
+    if len(r) > MAX_RANGE:
+        def limit():
+            i = 0
+            for item in r:
+                i += 1
+                yield item
+                if i >= MAX_RANGE:
+                    break
+        return list(limit())
+    return list(r)
 
 
 def generate_lorem_ipsum(n=5, html=True, min=20, max=100):
@@ -310,12 +351,31 @@ except TypeError, e:
 del _test_singleton, _test_gen_bug
 
 
-def buffereater(f):
+def pformat(obj, verbose=False):
+    """
+    Prettyprint an object.  Either use the `pretty` library or the
+    builtin `pprint`.
+    """
+    try:
+        from pretty import pretty
+        return pretty(obj, verbose=verbose)
+    except ImportError:
+        from pprint import pformat
+        return pformat(obj)
+
+
+def buffereater(f, template_data=False):
     """
     Used by the python translator to capture output of substreams.
     (macros, filter sections etc)
     """
-    return lambda *a, **kw: capture_generator(f(*a, **kw))
+    def wrapped(*a, **kw):
+        __traceback_hide__ = True
+        rv = capture_generator(f(*a, **kw))
+        if template_data:
+            rv = TemplateData(rv)
+        return rv
+    return wrapped
 
 
 def empty_block(context):
@@ -326,78 +386,6 @@ def empty_block(context):
     if 0: yield None
 
 
-def fake_template_exception(exception, filename, lineno, source,
-                            context_or_env):
-    """
-    Raise an exception "in a template". Return a traceback
-    object. This is used for runtime debugging, not compile time.
-    """
-    # some traceback systems allow to skip frames
-    __traceback_hide__ = True
-
-    from jinja.datastructure import Context
-    if isinstance(context_or_env, Context):
-        env = context_or_env.environment
-        namespace = context_or_env.to_dict()
-    else:
-        env = context_or_env
-        namespace = {}
-
-    # generate an jinja unique filename used so that linecache
-    # gets data that doesn't interferes with other modules
-    if filename is None:
-        from random import randrange
-        vfilename = 'jinja://~%d' % randrange(0, 10000)
-        filename = '<string>'
-    else:
-        vfilename = 'jinja://%s' % filename
-
-    offset = '\n' * (lineno - 1)
-    code = compile(offset + 'raise __exception_to_raise__',
-                   vfilename or '<template>', 'exec')
-
-    loader = TracebackLoader(env, source, filename)
-    loader.update_linecache(vfilename)
-    globals = {
-        '__name__':                 vfilename,
-        '__file__':                 vfilename,
-        '__loader__':               loader,
-        '__exception_to_raise__':   exception
-    }
-    try:
-        exec code in globals, namespace
-    except:
-        return sys.exc_info()
-
-
-def translate_exception(template, exc_type, exc_value, traceback, context):
-    """
-    Translate an exception and return the new traceback.
-    """
-    error_line = traceback.tb_lineno
-    for code_line, tmpl_filename, tmpl_line in template._debug_info[::-1]:
-        if code_line <= error_line:
-            break
-    else:
-        # no debug symbol found. give up
-        return traceback
-
-    return fake_template_exception(exc_value, tmpl_filename, tmpl_line,
-                                   template._source, context)[2]
-
-
-def raise_syntax_error(exception, env, source=None):
-    """
-    This method raises an exception that includes more debugging
-    informations so that debugging works better. Unlike
-    `translate_exception` this method raises the exception with
-    the traceback.
-    """
-    exc_info = fake_template_exception(exception, exception.filename,
-                                       exception.lineno, source, env)
-    raise exc_info[0], exc_info[1], exc_info[2]
-
-
 def collect_translations(ast):
     """
     Collect all translatable strings for the given ast. The
@@ -409,25 +397,16 @@ def collect_translations(ast):
     result = []
     while todo:
         node = todo.pop()
-        if node.__class__ is Trans:
+        if node.__class__ is nodes.Trans:
             result.append((node.lineno, node.singular, node.plural))
-        elif node.__class__ is CallFunc and \
-             node.node.__class__ is Name and \
+        elif node.__class__ is nodes.CallExpression and \
+             node.node.__class__ is nodes.NameExpression and \
              node.node.name == '_':
-            if len(node.args) in (1, 3):
-                args = []
-                for arg in node.args:
-                    if not arg.__class__ is Const:
-                        break
-                    args.append(arg.value)
-                else:
-                    if len(args) == 1:
-                        singular = args[0]
-                        plural = None
-                    else:
-                        singular, plural, _ = args
-                    result.append((node.lineno, singular, plural))
-        todo.extend(node.getChildNodes())
+            if len(node.args) == 1 and not node.kwargs and not node.dyn_args \
+               and not node.dyn_kwargs and \
+               node.args[0].__class__ is nodes.ConstantExpression:
+                result.append((node.lineno, node.args[0].value, None))
+        todo.extend(node.get_child_nodes())
     result.sort(lambda a, b: cmp(a[0], b[0]))
     return result
 
@@ -445,8 +424,7 @@ class DebugHelper(object):
 
     def __call__(self, env, context):
         """Print a nice representation of the context."""
-        from pprint import pformat
-        return pformat(context.to_dict())
+        return pformat(context.to_dict(), verbose=True)
 
     def filters(self, env, context, builtins=True):
         """List the filters."""
@@ -492,49 +470,6 @@ class DebugHelper(object):
 debug_helper = object.__new__(DebugHelper)
 
 
-class TracebackLoader(object):
-    """
-    Fake importer that just returns the source of a template.
-    """
-
-    def __init__(self, environment, source, filename):
-        self.loader = environment.loader
-        self.source = source
-        self.filename = filename
-
-    def update_linecache(self, virtual_filename):
-        """
-        Hacky way to let traceback systems know about the
-        Jinja template sourcecode. Very hackish indeed.
-        """
-        # check for linecache, not every implementation of python
-        # might have such an module.
-        try:
-            from linecache import cache
-        except ImportError:
-            return
-        data = self.get_source(None)
-        cache[virtual_filename] = (
-            len(data),
-            None,
-            data.splitlines(True),
-            virtual_filename
-        )
-
-    def get_source(self, impname):
-        source = ''
-        if self.source is not None:
-            source = self.source
-        elif self.loader is not None:
-            try:
-                source = self.loader.get_source(self.filename)
-            except TemplateNotFound:
-                pass
-        if isinstance(source, unicode):
-            source = source.encode('utf-8')
-        return source
-
-
 class CacheDict(object):
     """
     A dict like object that stores a limited number of items and forgets
@@ -698,3 +633,12 @@ class CacheDict(object):
         rv._mapping = deepcopy(self._mapping)
         rv._queue = deepcopy(self._queue)
         return rv
+
+
+NAMESPACE = {
+    'range':            safe_range,
+    'debug':            debug_helper,
+    'lipsum':           generate_lorem_ipsum,
+    'watchchanges':     watch_changes,
+    'rendertemplate':   render_included
+}
